------------------------------------------------------------------------
-- 
-- 7-stelliger Frequenzzaehler bis 200 MHz
-- 
-- CPLD : XC9572-PC84-15
--
-- Autor : Markus Lemke, DL1DSN, www.dl1dsn.de
--
-- Version : 1.0
--
-- Datum : 08.01.2003
--
-- Beschreibung:
-- Das VHDL-File counter.vhd enthaelt einen kompletten
-- Frequenzzaehler zur Ansteuerung einer 7-stelligen 
-- LED-Anzeige mit folgenden Modulen:
--
--    - 7 Dezimalzaehler, davon 6x 4bit, 1x 2bit
--    - Tor- und Ablaufsteuerung
--    - Multiplexer 7 zu 1, 4 bit breit (realisiert mit TBUFs im I/O)
--    - Zeitbasisteiler, max. Quarzfrequenz 8.192 MHz
--    - BCD-zu-7-Segment Decoder
--    - Logik zur Unterdrueckung fuehrender Nullen auf der Anzeige
--    - Multiplex-Steuerung fuer 7 Digits
--
-- This file is freeware and may be copied and modifyed by everybody
------------------------------------------------------------------------


library ieee, work;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

entity counter is 
   port
   (
      clk     : in std_logic; -- Zeitbasistakt 
      rst_n   : in std_logic;
      
      clk_x   : in std_logic; -- Zaehltakt
      
      -- LCD-Display
      cnt0    : out std_logic_vector(3 downto 0); -- alle Counter-Ausgaenge bitweise extern verbunden
      cnt1    : out std_logic_vector(3 downto 0);
      cnt2    : out std_logic_vector(3 downto 0);
      cnt3    : out std_logic_vector(3 downto 0);
      cnt4    : out std_logic_vector(3 downto 0);
      cnt5    : out std_logic_vector(3 downto 0);
      cnt6    : out std_logic_vector(3 downto 0);
      
      cnt_temp : out std_logic; -- Erklaerung s.u.
      
      cnt_in  : in std_logic_vector(3 downto 0);
      
      led     : out std_logic_vector(6 downto 0); -- 7 Segmente
      digit   : out std_logic_vector(6 downto 0)  -- 7 Stellen
   );
end;

architecture behave of counter is

   -- BCD-zu-7-Segment Decoder Funktion -------------------------------------
   function bcd_to_7seg( cnt : std_logic_vector(3 downto 0))
      return std_logic_vector
   is
      variable res : std_logic_vector(6 downto 0);
   begin
      case cnt is
         when x"0" => res := "1111110";
         when x"1" => res := "0110000";
         when x"2" => res := "1101101";
         when x"3" => res := "1111001";
         when x"4" => res := "0110011";
         when x"5" => res := "1011011";
         when x"6" => res := "0011111";
         when x"7" => res := "1110000";
         when x"8" => res := "1111111";
         when x"9" => res := "1110011";
         when x"A" => res := "1110111"; -- Decodierung > 9 wird nicht benoetigt,
         when x"B" => res := "0011111"; -- kostet aber auch keine zusaetzlichen Ressourcen
         when x"C" => res := "1001110";
         when x"D" => res := "0111101";
         when x"E" => res := "1001111";
         when x"F" => res := "1000111";
         when others => res := "1111111"; -- nur fuer Simulation
      end case;
      return res;
   end;
   -- bei groesseren Designs wuerde man die Funktion in ein Package auslagern
   --------------------------------------------------------------------------
   
   constant DIGIT_DELAY : integer := 1; -- fuer 2 ms Leuchtdauer / Digit
   
   constant ZERO : std_logic_vector(cnt0'RANGE) := (others=>'0'); -- ZERO = "0000"
   
   constant ZEITBASIS : integer := 3999; -- = Quarzfrequenz der Zeitbasis in kHz -1 (1MHz -> 999, 4 MHz -> 3999)
   signal basis_cnt_q : integer range 0 to ZEITBASIS;
   
   -- Statemachine mit 10 States zur Steuerung des Frequenzzaehlers
   type ST_TYPE is (S_1, S_2, S_3, S_4, S_5, S_6, S_7, S_8, S_9, S_10);
   signal st_q : ST_TYPE;
   
   signal rst_cnt_q   : std_logic;
   signal open_gate_q : boolean;
   
   signal cnt0_q : std_logic_vector(3 downto 0); -- 4bit Dezimalzaehler
   signal cnt1_q : std_logic_vector(3 downto 0); 
   signal cnt2_q : std_logic_vector(3 downto 0);
   signal cnt3_q : std_logic_vector(3 downto 0);
   signal cnt4_q : std_logic_vector(3 downto 0);
   signal cnt5_q : std_logic_vector(3 downto 0);
   signal cnt6_q : std_logic_vector(1 downto 0); -- zaehlt ja max. nur bis 200 MHz, 2 bit reichen also (bei 199MHz auch 1bit)
   
   signal digit_cnt_q : integer range 0 to 1;
   signal time_cnt_q  : integer range 0 to 9;
   signal drive_q     : boolean; -- Hilfssignal zur Vornullen-Unterdrueckung
   signal loop_q      : std_logic_vector(2 downto 0); -- 0 .. 7
   signal basis_q     : boolean;
   
   -- Zwischen-Taktsignale, von den Ausgaengen des 1. und 2. Dezimalzaehlers abgeleitet
   signal clk_0 : std_logic;
   signal clk_1 : std_logic;
   
begin
   -- CLK von Zeitbasis, z.B. 1 MHz
   process(rst_n, clk)
      variable ld_digit_cnt : boolean;
   begin
      if rst_n='0' then
         basis_cnt_q <= ZEITBASIS;
         st_q        <= S_1;
         rst_cnt_q   <= '1';
         open_gate_q <= false;
         digit_cnt_q <= 0;
         basis_q         <= false;
         time_cnt_q  <= 0;
         drive_q     <= false;
         loop_q      <= (others=>'0');
         
      elsif falling_edge(clk) then -- fallende Flanke der Zeitbasis verwenden
         ld_digit_cnt := false;
      
         -- Asynchroner Reset fuer den Counter
         rst_cnt_q <= '0';
      
         -- Zeitbasis-Counter teilt die Zeitbasis / 1000
         if basis_cnt_q = 0 then
            basis_cnt_q <= ZEITBASIS;
         else
            basis_cnt_q <= basis_cnt_q -1;
         end if;
         basis_q <= basis_cnt_q = 0; -- erzeugt ein  1 ms-Zeitraster
         
         -- Statemachine zur Zaehler-Steuerung
         case st_q is
            when S_1 =>
               if basis_q then
                  open_gate_q <= loop_q=0; -- nur alle X Zyklen eine Messung durchfuehren!!!
                  time_cnt_q  <= 0;
                  st_q <= S_2;
               end if;
               drive_q <= false; -- Anzeige dunkel beim Zaehlen
               
            when S_2 =>
               if basis_q then
                  if time_cnt_q = 9 then
                     open_gate_q  <= false;
                     st_q         <= S_3;
                     ld_digit_cnt := true;
                  else
                     time_cnt_q <= time_cnt_q +1;
                  end if;
               end if;
            
            when S_3 =>
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  st_q     <= S_4;
                  loop_q   <= loop_q +1;
               end if;
               if cnt_in /= x"0" then
                  drive_q <= true; -- jetzt Anzeige aktivieren, falls Zahlerstelle > 0
               end if;

            when S_4 =>
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  st_q     <= S_5;
               end if;
               if cnt_in /= x"0" then
                  drive_q <= true;
               end if;
               
            when S_5 =>
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  st_q     <= S_6;
               end if;
               if cnt_in /= x"0" then
                  drive_q <= true;
               end if;
               
            when S_6 =>
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  st_q     <= S_7;
               end if;
               if cnt_in /= x"0" then
                  drive_q <= true;
               end if;
                  
            when S_7 =>
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  st_q     <= S_8;
               end if;
               if cnt_in /= x"0" then
                  drive_q <= true;
               end if;
                  
            when S_8 =>
               drive_q <= true;
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  st_q     <= S_9;
               end if;
               
            when S_9 =>
               if digit_cnt_q = 0 and basis_cnt_q = 1 then
                  digit_cnt_q  <= 0;
                  if loop_q = 0 then
                     rst_cnt_q <= '1';
                     st_q <= S_10;
                  else
                     st_q <= S_1;
                  end if;
               end if;
            
            when S_10 =>
               rst_cnt_q <= '0';
               st_q      <= S_1;
         end case;
         
         -- Hilfszaehler digit_cnt_q steuert die Leuchtdauer der einzelnen 7-Seg-Anzeigen
         -- mit DIGIT_DELAY=1 erhaelt man 2ms pro Stelle
         if ld_digit_cnt then
            digit_cnt_q <= DIGIT_DELAY;
         elsif basis_cnt_q = 0 then
            if digit_cnt_q = 0 then
               digit_cnt_q <= DIGIT_DELAY;
            else
               digit_cnt_q <= digit_cnt_q -1;
            end if;
         end if;
         
      end if;
   end process;

   -- der aktuelle Zustand von st_q steuert gleichzeitig den Zaehler-MUX
   cnt6 <= "00" & cnt6_q when st_q = S_3 else (others=>'Z'); -- 100 MHz-Stelle zuerst anzeigen
   cnt5 <= cnt5_q when st_q = S_4 else (others=>'Z');
   cnt4 <= cnt4_q when st_q = S_5 else (others=>'Z');
   cnt3 <= cnt3_q when st_q = S_6 else (others=>'Z');
   cnt2 <= cnt2_q when st_q = S_7 else (others=>'Z');
   cnt1 <= cnt1_q when st_q = S_8 else (others=>'Z');
   cnt0 <= cnt0_q when st_q = S_9 else (others=>'Z'); -- 100 Hz - Stelle zuletzt anzeigen

   -- dieser Ausgang wird nur benoetigt, damit alle FF's der 1. Zaehlerstufe im selben FunctionBlock
   -- implementiert werden (im Layout offener Pin!)
   -- dasselbe Signal wird fuer den MUX nochmals rausgefuehrt(an einen Pin, der fuer das Layout guenstiger ist)
   cnt_temp <= cnt0_q(3) when st_q = S_9 else 'Z';
   
   -- Multiplex-Steuerung
   -- Die Digit-Treiber sind pnp-Transistoren und benoetigen
   -- ein L-Signal, damit die entspr. 7-Seg-Anzeige leuchtet
   -- Achtung! die Treiber-Steuerung wird rein kombinatorisch aus 
   -- den Signalen drive_q (= Triber generell aktiviert) und st_q
   -- (Statemachine) abgeleitet
   process(st_q, drive_q)
   begin

      if drive_q then
         case st_q is
            when S_3 => digit <= "0111111"; -- 0 = Digit-Treiber aktiv
            when S_4 => digit <= "1011111";
            when S_5 => digit <= "1101111";
            when S_6 => digit <= "1110111";
            when S_7 => digit <= "1111011";
            when S_8 => digit <= "1111101";
            when S_9 => digit <= "1111110";
            when others => digit <= (others=>'1'); -- alles AUS
         end case;
      else
         digit <= (others=>'1');
      end if;
      
   end process;
   
   -- cnt_in ist praktisch der Ausgang des Tristate-Multiplexers,
   -- mit dem nacheinander die 7 Counter-Ausgaenge auf die Anzeige
   -- gemultiplext werden
   -- Die Funktion bcd_to_7seg wandelt das 4 bit-Signal cnt_in 
   -- in ein 7-bit Signal fuer die LED-Anzeige.
   led <= not bcd_to_7seg(cnt_in); -- Invertierung, da Anzeige mit gem. Anode verwendet wird

   -- Counter 0 -> wird direkt vom zu messenden Signal getaktet
   -- und hat damit die hoechsten Timing-Anforderungen
   -- Damit 200MHz Taktfrequenz erreicht werden, muessen alle 4 bit
   -- dieses Counters im selben Function-Block des CPLD liegen !!! (4 FB's gibt es insgesamt im 9572)
   -- Nur dann koennen fuer alle bits des Counters "Local Feedback Signale"
   -- verwendet werden. Ansonsten gehen die Ausgange des cnt0_q zur Globalen Switch-Matrix im CPLD
   -- -> wesentlich langsamer !
   -- Weiterhin muessen die verwendeten Macrozellen im sog. "High Performance Mode" konfiguriert sein
   process(rst_cnt_q, clk_x)
   begin
      if rst_cnt_q='1' then
         cnt0_q <= ZERO; -- asynchroner Reset fuer Counter cnt0_q = "0000"
         
      elsif falling_edge(clk_x) then -- fallende Flanke ist bei Vorverstrkern meist steiler
      
         if open_gate_q then -- Tor geoeffnet ?
            -- 1. Stelle (100 Hz)
            if cnt0_q = 9 then
               cnt0_q <= ZERO;
            else
               cnt0_q <= cnt0_q +1;
            end if;
         end if;
         
      end if;
   end process;
   clk_0 <= cnt0_q(3); -- Clock fuer naechste Stelle wird vom MSB abgeleitet

   -- Counter 1 -> erhaelt den Takt vom MSB des Counter 0
   -- und hat damit die zweithoechsten Timing-Anforderungen 
   -- Macrozellen fuer diesen Counter muessen ebenfalls wie fuer Counter 0
   -- im High-Performance-Mode arbeiten
   process(rst_cnt_q, clk_0)
   begin
      if rst_cnt_q='1' then
         cnt1_q <= ZERO;
         
      elsif falling_edge(clk_0) then

         -- 2. Stelle (1 kHz)
         if cnt1_q = 9 then
            cnt1_q <= ZERO;
         else
            cnt1_q <= cnt1_q +1;
         end if;

      end if;
   end process;
   clk_1 <= cnt1_q(3); -- Clock fuer naechste Stelle wird vom MSB abgeleitet
   
   -- Alle folgenden Stellen (ab der 3. Stelle) arbeiten als Synchronzaehler mit clk_1 als Takt,
   -- d.h. alle Flip-Flops der 3. - 7. Zaehlerstelle bekommen den selben Takt
   process(rst_cnt_q, clk_1)
      variable cy2 : boolean; -- Carry-Signale
      variable cy3 : boolean;
      variable cy4 : boolean;
      variable cy5 : boolean;
   begin
      if rst_cnt_q='1' then
         cnt2_q <= ZERO;
         cnt3_q <= ZERO;
         cnt4_q <= ZERO;
         cnt5_q <= ZERO;
         cnt6_q <= "00";
         
      elsif falling_edge(clk_1) then
         
         -- Carry-Signale der einzelnen Dezimal-Zaehler generieren
         -- Uebertrag auf jeweils naechste Stelle, wenn die
         -- vorhergehende Stelle auf 9 steht
         cy2 :=         cnt2_q = 9;
         cy3 := cy2 and cnt3_q = 9;
         cy4 := cy3 and cnt4_q = 9;
         cy5 := cy4 and cnt5_q = 9;
         
         -- 3. Stelle (10 kHz)
         if cnt2_q = 9 then
            cnt2_q <= ZERO;
         else
            cnt2_q <= cnt2_q +1;
         end if;
         
         -- 4. Stelle (100 kHz)
         if cy2 then
            if cnt3_q = 9 then
               cnt3_q <= ZERO;
            else
               cnt3_q <= cnt3_q +1;
            end if;
         end if;
         
         -- 5. Stelle (1 MHz)
         if cy3 then
            if cnt4_q = 9 then
               cnt4_q <= ZERO;
            else 
               cnt4_q <= cnt4_q +1;
            end if;
         end if;

         -- 6. Stelle (10 MHz)
         if cy4 then
            if cnt5_q = 9 then
               cnt5_q <= ZERO;
            else 
               cnt5_q <= cnt5_q +1;
            end if;
         end if;
         
         -- 7. Stelle (100 MHz)
         if cy5 then -- kommt max nur bis 2, keine Abfrage auf 9 notwendig!
            cnt6_q <= cnt6_q +1;
         end if;
         
      end if;
   end process;
end;




